/* FILE: atlasscalarfield.cc      -*-Mode: c++-*-
 *
 * Atlas scalar field object, derived from Oxs_ScalarField
 * class.
 *
 */

#include <string>

#include "oc.h"
#include "nb.h"

#include "util.h"
#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "atlas.h"
#include "atlasscalarfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

OC_USE_STRING;

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_AtlasScalarField);

/* End includes */

// Constructor
Oxs_AtlasScalarField::Oxs_AtlasScalarField(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_ScalarField(name,newdtr,argstr),
    multiplier(1.0)
{
  // Process arguments
  OXS_GET_INIT_EXT_OBJECT("atlas",Oxs_Atlas,atlas_obj);
  atlas_key.Set(atlas_obj.GetPtr());
  atlas_key.GetReadReference(); // Read lock held until *this is deleted.

  multiplier = GetRealInitValue("multiplier",1.0);

  // Optional "default_value"
  if(HasInitValue("default_value")) {
    OXS_GET_INIT_EXT_OBJECT("default_value",Oxs_ScalarField,
                            default_value.value_init);
    default_value.value_is_set=1;
  }

  // Set up value array.
  values.SetSize(static_cast<OC_INDEX>(atlas_obj->GetRegionCount()));

  // Fill value vector with non-default values
  size_t i;
  vector<String> params;
  FindRequiredInitValue("values",params);
  if(params.size()%2!=0) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"List associated with parameter \"values\" has"
		" %u entries, which is not an even number.",
		(unsigned int)params.size());
    throw Oxs_ExtError(this,buf);
  }
  for(i=0;i<params.size();i+=2) {
    const String& region_name = params[i];
    OC_INDEX region_id
      = static_cast<OC_INDEX>(atlas_obj->GetRegionId(region_name));
    if(region_id<0) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Specified name, \"%s\","
		  " is not a known region in atlas \"%s\"."
		  "  Known regions:",
		  region_name.c_str(),atlas_obj->InstanceName());
      String msg = String(buf);
      vector<String> regions;
      atlas_obj->GetRegionList(regions);
      for(unsigned int j=0;j<regions.size();++j) {
	msg += String("\n ");
	msg += regions[j];
      }
      throw Oxs_ExtError(this,msg);
    }
    if(values[region_id].value_is_set) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Region \"%s\" in atlas \"%s\" is set more than once.",
		  region_name.c_str(),atlas_obj->InstanceName());
      throw Oxs_ExtError(this,buf);
    }
    Oxs_SplitList field_paramlist;
    vector<String> field_params;
    field_paramlist.Split(params[i+1]);
    field_paramlist.FillParams(field_params);
    OXS_GET_EXT_OBJECT(field_params,Oxs_ScalarField,
                       values[region_id].value_init);
    values[region_id].value_is_set = 1;
  }
  DeleteInitValue("values");

  VerifyAllInitArgsUsed();
}

OC_REAL8m
Oxs_AtlasScalarField::Value(const ThreeVector& pt) const
{
  OC_REAL8m result=0.0;
  const Oxs_Atlas* atlas = atlas_key.GetPtr();
  OC_INT4m id = atlas->GetRegionId(pt);
  if(values[OC_INDEX(id)].value_is_set) {
    result = values[OC_INDEX(id)].value_init->Value(pt);
  } else if(default_value.value_is_set) {
    result = default_value.value_init->Value(pt);
  } else {
    char buf[1024];
    String regionname;
    atlas->GetRegionName(id,regionname);
    Oc_Snprintf(buf,sizeof(buf),
		"Error in function Value():"
		" input point (%g,%g,%g), in region %s of atlas %s,"
		" has not been assigned a value.",
		static_cast<double>(pt.x),
                static_cast<double>(pt.y),
                static_cast<double>(pt.z),
		regionname.c_str(),atlas->InstanceName());
    throw Oxs_ExtError(this,buf);
  }
  return result*multiplier;
}
